from __future__ import annotations

import inspect
import typing
import warnings
from typing import Any, Dict, Optional, cast, no_type_check

from extendable import context, main
from extendable.main import ExtendableMeta
from extendable.registry import ExtendableClassesRegistry, ExtendableRegistryListener

try:
    from typing import _TypingBase  # type: ignore[attr-defined,unused-ignore]
except ImportError:
    from typing import _Final as _TypingBase  # type: ignore[attr-defined,unused-ignore]

from pydantic._internal._model_construction import ModelMetaclass
from pydantic.main import BaseModel, FieldInfo

from .utils import all_identical, resolve_annotation

typing_base = _TypingBase


if typing.TYPE_CHECKING:
    AnyClassmethod = classmethod[Any, Any, Any]


class ExtendableModelMeta(ExtendableMeta, ModelMetaclass):
    @no_type_check
    @classmethod
    def _build_original_class(metacls, name, bases, namespace, **kwargs):
        if not main._registry_build_mode and BaseModel in bases:
            # we must wrap all the classmethod defined into pydantic.BaseModel
            metacls._wrap_pydantic_base_model_class_methods(namespace)
            pass
        return ModelMetaclass.__new__(metacls, name, bases, namespace, **kwargs)

    @no_type_check
    @classmethod
    def _prepare_namespace(
        metacls, name, bases, namespace, extends=None, **kwargs
    ) -> Dict[str, Any]:
        namespace = super()._prepare_namespace(
            name=name, bases=bases, namespace=namespace, extends=extends, **kwargs
        )
        return namespace

    @no_type_check
    def __call__(cls, *args, **kwargs) -> "ExtendableMeta":
        """Create the aggregated class in place of the original class definition.

        This method called at instance creation. The resulted instance
        will be an instance of the aggregated class not an instance of
        the original class definition since this definition could have
        been extended.
        """
        is_aggregated = getattr(
            cls._is_aggregated_class, "default", cls._is_aggregated_class
        )
        if is_aggregated:
            return super().__call__(*args, **kwargs)
        return cls._get_assembled_cls()(*args, **kwargs)

    @classmethod
    def _wrap_pydantic_base_model_class_methods(
        metacls, namespace: Dict[str, Any]
    ) -> Dict[str, Any]:
        new_namespace = namespace
        from pydantic.warnings import PydanticDeprecationWarning

        with warnings.catch_warnings():
            # ignore warnings about deprecated methods into pydantic.BaseModel
            warnings.filterwarnings("ignore", category=PydanticDeprecationWarning)
            methods = inspect.getmembers(BaseModel, inspect.ismethod)
            for name, method in methods:
                if name.startswith("__"):
                    continue
                if name in namespace:
                    continue
                new_namespace[name] = metacls._wrap_class_method(
                    cast("AnyClassmethod", method), name
                )
        return new_namespace

    ###############################################################
    # concrete methods provided to the final class by the metaclass
    ###############################################################
    def _resolve_submodel_fields(
        cls, registry: Optional[ExtendableClassesRegistry] = None
    ) -> None:
        """Replace the original field type into the definition of the field by the one
        from the registry."""
        registry = registry if registry else context.extendable_registry.get()
        to_rebuild = False
        if issubclass(cls, BaseModel):
            type_namespace = cast(BaseModel, cls).__pydantic_parent_namespace__ or {}
            name_space = type_namespace.get("namespace", {})
            for field_name, annoted_type in name_space.get(
                "__annotations__", {}
            ).items():
                if issubclass(type(annoted_type), ExtendableModelMeta):
                    name_space["__annotations__"][
                        field_name
                    ] = annoted_type._get_assembled_cls(registry)
                    to_rebuild = True
            for field_name, field_info in cast(BaseModel, cls).model_fields.items():
                new_type = resolve_annotation(field_info.annotation, registry)
                if not all_identical(field_info.annotation, new_type):
                    cast(BaseModel, cls).model_fields[
                        field_name
                    ] = FieldInfo.from_annotation(new_type)
                    to_rebuild = True
        if to_rebuild:
            delattr(cls, "__pydantic_core_schema__")
            cast(BaseModel, cls).model_rebuild(force=True)
            return


class RegistryListener(ExtendableRegistryListener):
    def on_registry_initialized(self, registry: ExtendableClassesRegistry) -> None:
        self.resolve_submodel_fields(registry)
        self.rebuild_models(registry)

    def rebuild_models(self, registry: ExtendableClassesRegistry) -> None:
        for cls in registry._extendable_classes.values():
            cast(BaseModel, cls).model_rebuild(force=True)

    def resolve_submodel_fields(self, registry: ExtendableClassesRegistry) -> None:
        for cls in registry._extendable_classes.values():
            if issubclass(type(cls), ExtendableModelMeta):
                cast(ExtendableModelMeta, cls)._resolve_submodel_fields(registry)


ExtendableClassesRegistry.listeners.append(RegistryListener())

from pydantic._internal import _generate_schema  # noqa: E402

initial_type_ref = _generate_schema.get_type_ref


def get_type_ref(
    type_: type[Any], args_override: tuple[type[Any], ...] | None = None
) -> str:
    type_ref: str = initial_type_ref(type_, args_override)
    # Ensure type_ref unicity for each extendable model
    if issubclass(type(type_), ExtendableModelMeta):
        module_name = getattr(type_, "__module__", "<No __module__>")
        type_ref = f"{module_name}.{type_.__name__}:{id(type_)}"
    return type_ref


_generate_schema.get_type_ref = get_type_ref



#  type: ignore
# ruff: noqa: E402

# The following 2 lines are required to prevent import error if others dependecies
# are not yet available into the python path when this file is loaded.
# This should be done by default unless the Python interpreter was started
# with the -S flag. It's at least required when this lib is deployed into odoo.sh
import site

site.main()


try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated

import wrapt
from extendable import context
from pydantic import TypeAdapter

from .utils import all_identical, resolve_annotation


def _resolve_model_fields_annotation(model_fields):
    registry = context.extendable_registry.get()
    if registry:
        for field in model_fields:
            field_info = field.field_info
            new_type = resolve_annotation(field_info.annotation)
            if not all_identical(field_info.annotation, new_type):
                field_info.annotation = new_type
                field._type_adapter = TypeAdapter(Annotated[new_type, field_info])
    return model_fields


@wrapt.when_imported("fastapi.openapi.utils")
def hook_fastapi_openapi_utils(utils):
    # This method is used by fastapi to build the fields definition for all
    # the routes. We need to replace it with a wrapper that will replace
    # in the annotation any class that has been extended with the extended
    # class.
    # These definitions are used by fastapi to build the openapi schema.
    def _get_fields_from_routes_wrapper(wrapped, instance, args, kwargs):
        all_fields = wrapped(*args, **kwargs)
        _resolve_model_fields_annotation(all_fields)
        return all_fields

    wrapt.wrap_function_wrapper(
        utils, "get_fields_from_routes", _get_fields_from_routes_wrapper
    )


@wrapt.when_imported("fastapi.utils")
def hook_fastapi_utils(utils):
    # This method is used by fastapi to build the fields definition for all
    # the routes used to parse the request and to build the response.
    # This method is called a first time when the modules where the routes
    # are defined are imported. At this time the registry is not yet
    # initialized.
    # This method is then called every time a route is added to an app.
    # At this time the registry must be initialized. (it's the case
    # with the odoo.addons.fastapi module)
    def _create_response_field_wrapper(wrapped, instance, args, kwargs):
        field = wrapped(*args, **kwargs)
        _resolve_model_fields_annotation([field])
        return field

    wrapt.wrap_function_wrapper(
        utils, "create_response_field", _create_response_field_wrapper
    )



from __future__ import annotations

import sys
import types
import typing
from itertools import zip_longest
from typing import Any, List, Optional

import typing_extensions
from extendable import context
from extendable.main import ExtendableMeta
from extendable.registry import ExtendableClassesRegistry
from typing_extensions import get_args, get_origin

try:
    from typing import _TypingBase  # type: ignore[attr-defined,unused-ignore]
except ImportError:
    from typing import _Final as _TypingBase  # type: ignore[attr-defined,unused-ignore]

if sys.version_info >= (3, 10):
    from typing import _UnionGenericAlias  # type: ignore[attr-defined,unused-ignore]


typing_base = _TypingBase
_EMPTY = object()


def all_identical(left: Optional[Any], right: Optional[Any]) -> bool:
    """Check that the items of `left` are the same objects as those in `right`.

    >>> a, b = object(), object()
    >>> all_identical([a, b, a], [a, b, a])
    True
    >>> all_identical([a, b, [a]], [a, b, [a]])  # new list object, while "equal" is not "identical"
    False

    Credit: Pydantic
    """
    # check that left and right are iterables
    if not isinstance(left, typing.Iterable) or not isinstance(right, typing.Iterable):
        return left is right
    for left_item, right_item in zip_longest(left, right, fillvalue=_EMPTY):
        if left_item is not right_item:
            return False
    return True


def resolve_annotation(
    type_: Any, registry: Optional[ExtendableClassesRegistry] = None
) -> Any:
    """Return type with all occurrences of subclass of `ExtendableModelMeta` keys
    recursively replaced with their assembled class.

    Args:
        type_: The class or generic alias.
        registry:

    Returns:
        A new type representing the basic structure of `type_`

        Example:
    ```py
    from typing import List, Tuple, Union

    replace_types(Tuple[str, Union[List[MyExtendale], float]], registry)
    #> Tuple[int, Union[List[MyExtendale._get_assembled_cls(registry)], float]]
    ```

    Credits: Inspired from pydantic._internal._generics.replace_types

    see (https://github.com/pydantic/pydantic/blob/main/pydantic/_internal/_generics.py#L257C5-L257C18)
    """
    registry = registry if registry else context.extendable_registry.get()

    type_args = get_args(type_)
    origin_type = get_origin(type_)

    if origin_type is typing_extensions.Annotated:
        annotated_type, *annotations = type_args
        annotated = resolve_annotation(annotated_type, registry)
        for annotation in annotations:
            annotated = typing_extensions.Annotated[annotated, annotation]
        return annotated

    # Having type args is a good indicator that this is a typing module
    # class instantiation or a generic alias of some sort.
    if type_args:
        resolved_type_args = tuple(
            resolve_annotation(arg, registry) for arg in type_args
        )
        if all_identical(type_args, resolved_type_args):
            # If all arguments are the same, there is no need to modify the
            # type or create a new object at all
            return type_
        if (
            origin_type is not None
            and isinstance(type_, typing_base)
            and not isinstance(origin_type, typing_base)
            and getattr(type_, "_name", None) is not None
        ):
            # In python < 3.9 generic aliases don't exist so any of these like `list`,
            # `type` or `collections.abc.Callable` need to be translated.
            # See: https://www.python.org/dev/peps/pep-0585
            origin_type = getattr(typing, type_._name)
        assert origin_type is not None
        # PEP-604 syntax (Ex.: list | str) is represented with a types.UnionType object that does not have __getitem__.
        # We also cannot use isinstance() since we have to compare types.
        if sys.version_info >= (3, 10) and origin_type is types.UnionType:  # noqa: E721
            return _UnionGenericAlias(origin_type, resolved_type_args)
        return origin_type[resolved_type_args]

    # We handle pydantic generic models separately as they don't have the same
    # semantics as "typing" classes or generic aliases

    if not origin_type and issubclass(type(type_), ExtendableMeta):
        return type_._get_assembled_cls(registry)

    # Handle special case for typehints that can have lists as arguments.
    # `typing.Callable[[int, str], int]` is an example for this.
    if isinstance(type_, (List, list)):
        resolved_list = [resolve_annotation(element, registry) for element in type_]
        if all_identical(type_, resolved_list):
            return type_
        return resolved_list

    return type_
